# encoding: utf-8
# package: qtils
# author: Daniel Kovacs <github.com/neonihil>
# file-version: 1.0
# licence: LGPL-v3

# -------------------------------------------------------------------------------
# Licence
# -------------------------------------------------------------------------------
# This file is part of Ulra Light Weight Qtils.
#
# Ulra Light Weight Qtils is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# Ulra Light Weight Qtils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ulra Light Weight Qtils. If not, see <https://www.gnu.org/licenses/lgpl>.
# -------------------------------------------------------------------------------


"""
Logging Utilities module
=============================


Logging enhancements


Attributes:
    
    LOG_FORMATS: Pre-defined logging formats for convenience.

        Available formats:

        =========================== ===================================================================================
        Name                        Example
        =========================== ===================================================================================
        ``LOG_FORMATS.SHORT``       ``2019-10-10 11:11:11   INFO    mypackage.mymodule.MyClass:      Hello World``
        ``LOG_FORMATS.THREAD``      ``2019-10-10 11:11:11   INFO    [tid:11 (myThread)]    
                                    mypackage.mymodule.MyClass:      Hello World``   
        ``LOG_FORMATS.PROCESS``     ``2019-10-10 11:11:11   INFO    [pid:777]   mypackage.mymodule.
                                    MyClass:      Hello World``
        ``LOG_FORMATS.FULL``        ``2019-10-10 11:11:11   INFO    [pid:777 tid:11(myThread)]  mypackage.
                                    mymodule.MyClass:      Hello World``
        =========================== ===================================================================================


"""



# -------------------------------------------------------------------------------
# imports
# -------------------------------------------------------------------------------

import logging

from .collections import qlist, qdict


# -----------------------------------------------------------------------------
# exports
# -----------------------------------------------------------------------------

__all__ = qlist()


# -----------------------------------------------------------------------------
# LOG_FORMATS
# -----------------------------------------------------------------------------

__all__.append("LOG_FORMATS")
LOG_FORMATS = qdict(
    SHORT="%(asctime)s\t%(levelname)s\t%(name)s:\t%(message)s",
    THREAD="%(asctime)s\t%(levelname)s\t[tid:%(thread)x (%(threadName)s)]\t%(name)s:\t%(message)s",
    PROCESS="%(asctime)s\t%(levelname)s\t[pid:%(process)d]\t%(name)s:\t%(message)s",
    FULL="%(asctime)s\t%(levelname)s\t[pid:%(process)d tid:%(thread)x (%(threadName)s)]\t%(name)s:\t%(message)s",
)



# -----------------------------------------------------------------------------
# _create_class_logger
# -----------------------------------------------------------------------------

def _create_class_logger(cls, channel: str = None, root_channel: str = None, attr_name: str = "__logger"):
    """Create and assign a logger to a class object
    """
    channel = channel or cls.__name__
    root_channel = root_channel if root_channel is not None else cls.__module__
    if root_channel:
        channel = root_channel + '.' + channel
    if attr_name.startswith('__'): 
        attr_name = '_' + cls.__name__ + attr_name
    setattr(cls, attr_name, logging.getLogger(channel))


# -----------------------------------------------------------------------------
# @logged
# -----------------------------------------------------------------------------

@__all__.register
def logged(*args, **kwargs):
    """Decorator to create and assign a logger to a class
    
    Arguments:
        channel (str): Name of the logger, defaults to ``cls.__name__``

        root_channel (str): Prefix to the name of the logger,
            helpful to namespace a large ammount of logged objects.

        attr_name (str): Name of the logger attribute. Defaults
            to ``__logger``. 

            .. note:: Variables begining with double underscore
                are class private variables, meaning they are
                accessible ONLY to methods within the *same* class, 
                and not to descendant classes or from the outside. Read
                more about it in :py:ref:`tut-private` documentation.


    Examples:
        
        Simplest use case, adding a logger to a class:

        >>> @logged
        ... class LoggedFoo():
        ...     def __init__(self):
        ...         self.__logger.info("Hello World!")
        ...
        >>> foo = LoggedFoo()
        >>> # INFO:LoggedFoo:Hello World!     
        >>> #
        >>> # doctest doesn't allow testing logging output, 
        >>> # so just believe me on this one. :) 
        >>> # And also, there are proper test cases in pytest.

        
        Loggers are class-private by default:

        >>> @logged
        ... class LoggedFoo():
        ...     def __init__(self):
        ...         self.__logger.info("Hello World from Foo!")
        ...
        >>> class DescendantFoo(LoggedFoo):
        ...     def __init__(self):
        ...         super().__init__()
        ...         # The following line will fail because
        ...         # it has no access to the parent classes
        ...         # class-private logger.
        ...         self.__logger.info("Hello World from Descendant!") # 
        ...
        >>> foo = DescendantFoo()
        Traceback (most recent call last):
        ...
        AttributeError: 'DescendantFoo' object has no attribute '_DescendantFoo__logger'


        The recommended solution is to create loggers for each subclass. This means 
        their channel will always reflect the class name where the log line is 
        comming from. This is helpful in a large project with complicated inheritance 
        trees. 

        >>> @logged
        ... class LoggedFoo():
        ...     def __init__(self):
        ...         self.__logger.info("Hello World from Foo!")
        ...
        >>> @logged
        ... class DescendantFoo(LoggedFoo):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.__logger.info("Hello World from Descendant!") # 
        ...
        >>> foo = DescendantFoo()
        >>> # INFO:LoggedFoo:Hello World from Foo!     
        >>> # INFO:DescendantFoo:Hello World from Descendant!
        >>> #
        >>> # Note the Logger Name difference for the two lines.


        The second solution is to create a logger with a non-private attribute name.
        This could be useful for situations when there is a bunch of small descendant
        classes which doesn't need their own logger. This typically happens in
        marshallers when each type has it's own little class for serializing/deserializing.


        >>> @logged(attr_name="logger")
        ... class LoggedFoo():
        ...     def __init__(self):
        ...         self.logger.info("Hello World from Foo!")
        ...
        >>> class DescendantFoo(LoggedFoo):
        ...     def __init__(self):
        ...         super().__init__()
        ...         self.logger.info("Hello World from Descendant!") # 
        ...
        >>> foo = DescendantFoo()
        >>> # INFO:LoggedFoo:Hello World from Foo!     
        >>> # INFO:LoggedFoo:Hello World from Descendant!
        >>>
        >>> # Both logs seem to come from the same line



    """
    if len(args) == 1 and isinstance(args[0], type):
        _create_class_logger(args[0])
        return args[0]
    def _logged(cls):
        _create_class_logger(cls, *args, **kwargs)
        return cls
    return _logged




# encoding: utf-8
# package: qtils
# author: Daniel Kovacs <github.com/neonihil>
# file-version: 1.0
# licence: LGPL-v3

# -------------------------------------------------------------------------------
# Licence
# -------------------------------------------------------------------------------
# This file is part of Ulra Light Weight Qtils.
#
# Ulra Light Weight Qtils is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# Ulra Light Weight Qtils is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with Ulra Light Weight Qtils. If not, see <https://www.gnu.org/licenses/lgpl>.
# -------------------------------------------------------------------------------


"""
See more usage examples in :ref:`tut_properties`

"""


# -------------------------------------------------------------------------------
# imports
# -------------------------------------------------------------------------------

import weakref

from .collections import qlist


# -----------------------------------------------------------------------------
# exports
# -----------------------------------------------------------------------------

__all__ = qlist()


# -----------------------------------------------------------------------------
# @weakproperty
# -----------------------------------------------------------------------------

@__all__.register
def weakproperty(setter):
    """Returns a property that stores values as :py:class:`weakref.Ref()`

    The weakref object is stored as ``'_' + setter.__name__``

    Args:
        setter (function): setter function (can be an empty function)

    Returns:
        returns property instance

    Example:

        >>> import sys
        >>> class SomeClass(object): pass
        >>> class Foo(object):
        ...     @weakproperty
        ...     def bar(self, value): pass
        ...
        >>> some_obj = SomeClass()
        >>> original_refcount = sys.getrefcount(some_obj)
        >>> foo = Foo()
        >>> foo.bar = some_obj  # Reference count should not increase because some_obj is weak referenced.
        >>> original_refcount == sys.getrefcount(some_obj)
        True
        >>> foo._bar # weakref object stored as private variable
        <weakref at ... to 'SomeClass' at ...>
        >>> foo.bar
        <qtils.properties.SomeClass object at ...>
        >>> foo.bar == some_obj
        True
    
    """
    name = "_" + setter.__name__
    def _getter(self):
        value = getattr(self, name, None)
        return value() if isinstance(value, weakref.ref) else value
    def _setter(self, value):
        ref = weakref.ref(value) if value is not None else None
        setattr(self, name, ref)
        setter(self, value)
    return property(_getter, _setter)


# -----------------------------------------------------------------------------
# cachedproperty
# -----------------------------------------------------------------------------    

@__all__.register
def cachedproperty(getter=None, setter=None, deleter=None, varname=None):
    """Returns a property that caches first return of getter until a del is called.

    Args:
        getter (function): Getter function
        setter (function): Setter function
        deleter (function): Deleter function
        varname (str): Variable name to store cached data at, defaults to ``getter.__name__``
    Returns:
        return (property): property object with caching ability

    Example:
        
        Tyical usage:

        >>> class Foo(object):
        ...     @cachedproperty
        ...     def bar(self):
        ...         # doing some super computation-intensive thing here
        ...         print('getter called')
        ...         return "hello world"
        ...
        >>> obj = Foo()
        >>> obj.bar         # first call, getter is called
        getter called
        'hello world'
        >>> obj.bar         # second call, getter is not called
        'hello world'
        >>> obj._bar        # data is stored under '_'+setter.__name__
        'hello world'
        >>> del obj.bar     # removing cached value
        >>> obj.bar         # getter is called again
        getter called
        'hello world'
        >>> del obj.bar
        >>> obj.bar = 'ni!' # value can also be set from the outside
        >>> obj.bar         # no getter will be called
        'ni!'

        Using a different variable to cache data:

        >>> class Foo(object):
        ...     @cachedproperty(varname='_some_var')
        ...     def bar(self):
        ...         return "result of intensive calculation"
        >>> obj = Foo()
        >>> obj.bar
        'result of intensive calculation'
        >>> obj._some_var   # cached data will be stored under custom name
        'result of intensive calculation'

    """
    varname_ = varname
    def _cachedproperty(getter):
        varname = varname_ or ('_' + getattr(getter, "__name__"))
        def _getter(self):
            value = getattr(self, varname, None)
            if value is None:
                value = getter(self)
                setattr(self, varname, value)
            return value
        def _setter(self, value):
            setattr(self, varname, value)
        def _deleter(self):
            setattr(self, varname, None)
        return property(_getter, setter or _setter, deleter or _deleter)
    if getter:
        return _cachedproperty(getter)
    return _cachedproperty






#!/usr/bin/env python
# package: qtils
# licence: MIT <https://opensource.org/licenses/MIT>
# author: Daniel Kovacs <mondomhogynincsen@gmail.com>
# file: qtils/setup.py
# file-version: 2.2.1


# ---------------------------------------------------------------------------------------
# configuration
# ---------------------------------------------------------------------------------------

NAME = "qtils"
VERSION = "0.11.0"
DESCRIPTION = """Qtils - pronounces as `cuteels` - is a syntactic sugar library to make sweet Python coding even sweeter."""
AUTHOR = "Daniel Kovacs <@neonihil>"
AUTHOR_EMAIL = "mondomhogynincsen@gmail.com"
MAINTAINER = "Daniel Kovacs"
MAINTAINER_EMAIL = "mondomhogynincsen@gmail.com"
SCM_URL= "https://github.com/ultralightweight/qtils"
KEYWORDS = []
CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Topic :: Software Development :: Libraries",
]
LICENCE="LGPLv3"


# ---------------------------------------------------------------------------------------
# imports
# ---------------------------------------------------------------------------------------

import codecs
import os
import re

from setuptools import setup, find_packages
try: # for pip >= 10
    from pip._internal.req import parse_requirements
except ImportError: # for pip <= 9.0.3
    from pip.req import parse_requirements


# ---------------------------------------------------------------------------------------
# _read()
# ---------------------------------------------------------------------------------------

def _read(*parts):
    with codecs.open(os.path.join(HOME, *parts), "rb", "utf-8") as f:
        return f.read()


# ---------------------------------------------------------------------------------------
# get_requirements
# ---------------------------------------------------------------------------------------

def get_requirements():
    packages, dependencies = [], []
    for ir in parse_requirements(os.path.join( HOME, 'requirements.txt' ), session=False):
        if ir.link:
            dependencies.append(ir.link.url)
            continue
        packages.append(str(ir.req))
    return packages, dependencies


# ---------------------------------------------------------------------------------------
# internal variables
# ---------------------------------------------------------------------------------------

HOME = os.path.abspath(os.path.dirname(__file__))
PACKAGES = find_packages(where='src')
INSTALL_REQUIRES, DEPENDENCY_LINKS = get_requirements()


# ---------------------------------------------------------------------------------------
# setup()
# ---------------------------------------------------------------------------------------

if __name__ == "__main__":
    setup(
        name=NAME,
        description=DESCRIPTION,
        license=LICENCE,
        url=SCM_URL,
        version=VERSION,
        author=AUTHOR,
        author_email=AUTHOR_EMAIL,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        keywords=KEYWORDS,
        long_description=_read("README.rst"),
        packages=PACKAGES,
        package_dir={"": "src"},
        zip_safe=False,
        classifiers=CLASSIFIERS,
        install_requires=INSTALL_REQUIRES,
        dependency_links=DEPENDENCY_LINKS,
        setup_requires=[
        ],
        tests_require=[
            'pytest',
        ],
    )
